module EL
  module TagFactory
  
    # methods allowing to generate HTML tags using plain Ruby
    # 
    # @example
    #   div_tag 'some <evil> string' #=> <div>some &lt;evil&gt; string</div>
    #
    # tag content will be escaped by default.
    # to emit content as is, use bang methods!
    #
    # @example
    #   div_tag! 'some <evil> string' #=> <div>some <evil> string</div>
    #
    # also you'll want to use bang methods when you have some nested tags
    # 
    # @example
    #   div_tag! do
    #     b_tag 'some <evil> string'
    #   end
    #   #=> <div><b>some &lt;evil&gt; string</b></div>
    #
    EConstants::GENERIC_TAGS.each do |tag|
      define_method tag + '_tag' do |*args, &proc|
        content, attrs = EUtils.html_tag_requisites(true, *args)
        '<%s%s>%s</%s>' % [tag, attrs, (proc ? CGI.escapeHTML(proc.call.to_s) : content), tag]
      end
      define_method tag + '_tag!' do |*args, &proc|
        content, attrs = EUtils.html_tag_requisites(false, *args)
        '<%s%s>%s</%s>' % [tag, attrs, (proc ? proc.call : content), tag]
      end
    end

    EConstants::EMPTY_TAGS.each do |tag|
      define_method tag + '_tag' do |*args|
        '<%s%s>' % [tag, EUtils.html_tag_requisites(true, *args).last]
      end
    end
    
    # shorthand for `script_tag`
    # so you can type `js_tag "file"` instead of
    # `script_tag type: 'text/javascript', src: "file.js"`
    #
    # if URL given as first argument, it should not contain extension.
    # also URL will be automatically modified if used within assets mapper.
    #
    # to set an URL that wont be modified in any way, use :src option.
    #
    def js_tag src, attrs = {}
      src.is_a?(Hash) && (attrs = src) && (src = nil)
      src = src ? assets_url(src) : attrs[:src]
      suffix = attrs.delete(:suffix)
      attrs[:src] = '%s.js%s' % [src, suffix]
      attrs[:type] ||= 'text/javascript'
      '<script%s></script>' % EUtils.html_tag_attrs(attrs)
    end

    # shorthand for `link_tag`
    # so you can type `css_tag "file"` instead of
    # `link_tag rel: 'stylesheet', href: "file.css"`
    # 
    # URL are handled exactly as per `js_tag`
    def css_tag src, attrs = {}
      src.is_a?(Hash) && (attrs = src) && (src = nil)
      src = src ? assets_url(src) : (attrs[:href] || attrs.delete(:src))
      suffix = attrs.delete(:suffix)
      attrs[:href] = '%s.css%s' % [src, suffix]
      attrs[:media] ||= 'all'
       attrs[:type] ||= 'text/css'
        attrs[:rel] ||= 'stylesheet'
      '<link%s>' % EUtils.html_tag_attrs(attrs)
    end

    # shorthands for `img_tag`
    # so you can type `png_tag "file"` instead of `img_tag "file.png"`
    #
    # URL are handled exactly as per `js_tag`
    EConstants::IMAGE_TAGS.each do |ext|
      pattern = '%s.' + ext + '%s'
      define_method(ext + '_tag') do |src, attrs = {}|
        src.is_a?(Hash) && (attrs = src) && (src = nil)
        src = src ? assets_url(src) : attrs[:src]
        suffix = attrs.delete(:suffix)
        attrs[:src] = pattern % [src, suffix]
        '<img%s>' % EUtils.html_tag_attrs(attrs)
      end
    end

    def doctype_tag &proc
      "<!DOCTYPE html>\n%s" % (proc ? proc.call : '')
    end

    def comment_tag comment = nil
      comment = yield if block_given?
      '<!-- %s -->' % CGI.escapeHTML(comment.to_s)
    end

    # same as `comment_tag` except it wont escape comment!
    def comment_tag! comment = nil
      comment = yield if block_given?
      '<!-- %s -->' % comment
    end

    private

    # just a placeholder to work when used outside Espresso
    def assets_url src
      src
    end

  end
end

class E
  include EL::TagFactory

  # build a HTML <a> tag.
  #
  # if first param is a valid action, the URL of given action will be used.
  # action accepted as a symbol or a string representing action name and format.
  # action can also be passed in deRESTified form, eg. :read instead of :post_read
  #
  # @example
  #
  #   class App < E
  #     format '.html'
  #
  #     def read
  #       link_to :read        #=> /app/read
  #       link_to 'read.html'  #=> /app/read.html
  #       link_to 'read.xml'   #=> read.xml - not translated, used as is
  #     end
  #
  #     def post_write
  #       link_to :post_write  #=> /app/write - works but it is tedious, use :write instead
  #       link_to :write       #=> /app/write
  #       link_to 'write.html' #=> /app/write.html
  #       link_to '/something' #=> /something - not translated, used as is
  #     end
  #   end
  #
  # if `nil` passed as first argument, a void link will be created
  #
  # @example
  #
  #   link_to nil, 'something' #=> <a href="javascript:void(null);">something</a>
  #
  # anchor can be passed via second argument.
  # if it is missing, the link will be used as anchor
  # @note the anchor will be HTML escaped
  # 
  # @example
  #
  #   link_to :something   #=> <a href="/something">/something</a>
  #   link_to :foo, 'bar'  #=> <a href="/foo">bar</a>
  # 
  # anchor can also be passed as a block
  # 
  # @example
  # 
  #   link_to(:foo) { 'bar' }  #=> <a href="/foo">bar</a>
  # 
  # attributes can be passed as a hash via last argument
  # 
  # @example
  # 
  #   link_to :foo, target: '_blank'        #=> <a href="/foo" target="_blank">/foo</a>
  #   link_to :foo, :bar, target: '_blank'  #=> <a href="/foo" target="_blank">bar</a>
  # 
  def link_to *args, &proc
    '<a href="%s"%s>%s</a>' %
      EUtils.requisites_for_link_to(true, self.class, *args, &proc)
  end

  # same as `link_to` except it wont escape the anchor
  def link_to! *args, &proc
    '<a href="%s"%s>%s</a>' %
      EUtils.requisites_for_link_to(false, self.class, *args, &proc)
  end
end
